library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

entity opb_xsb300 is
  generic (
    C_OPB_AWIDTH    : integer := 32;
    C_OPB_DWIDTH    : integer := 32;
    C_BASEADDR      : std_logic_vector := X"2000_0000";
    C_HIGHADDR      : std_logic_vector := X"2000_00FF"
    );

    Port ( 	OPB_Clk : in std_logic;
				OPB_Rst : in std_logic;
				OPB_ABus : in std_logic_vector (31 downto 0);
				OPB_BE : in std_logic_vector (3 downto 0);
				OPB_DBus : in std_logic_vector (31 downto 0);
				OPB_RNW : in std_logic;
				OPB_select : in std_logic;
				OPB_seqAddr : in std_logic;
				pixel_clock : in std_logic;
				UIO_DBus : out std_logic_vector (31 downto 0);
				UIO_errAck : out std_logic;
				UIO_retry : out std_logic;
				UIO_toutSup : out std_logic;
				UIO_xferAck : out std_logic;
				PB_A : out std_logic_vector (19 downto 0);
				PB_UB_N : out std_logic; 
				PB_LB_N : out std_logic;
				PB_WE_N : out std_logic;
				PB_OE_N : out std_logic;
				RAM_CE_N : out std_logic;
				VIDOUT_CLK : out std_logic;
				VIDOUT_RCR : out std_logic_vector (9 downto 0);
				VIDOUT_GY : out std_logic_vector (9 downto 0);
				VIDOUT_BCB : out std_logic_vector (9 downto 0);
				VIDOUT_BLANK_N : out std_logic;
				VIDOUT_HSYNC_N : out std_logic;
				VIDOUT_VSYNC_N : out std_logic;
				PB_D : inout std_logic_vector (15 downto 0)
	 		   );
end opb_xsb300;

architecture Behavioral of opb_xsb300 is

constant C_MASK : integer := 0; -- huge address window as we are a bridge


signal addr_mux : std_logic_vector(19 downto 0);
signal video_addr : std_logic_vector (19 downto 0);
signal video_data : std_logic_vector (15 downto 0);
signal video_req : std_logic;
signal video_ce : std_logic;
signal i : integer;
signal cs : std_logic;

signal onecycle : std_logic ;
signal videocycle, amuxsel, hihalf : std_logic;
signal rce0, rce1, rreset : std_logic;
signal xfer : std_logic;
signal pb_wr, pb_rd : std_logic;

signal sram_ce : std_logic;

signal rnw : std_logic;

signal addr : std_logic_vector (23 downto 0); 	

signal be : std_logic_vector (3 downto 0);
signal pb_bytesel : std_logic_vector (1 downto 0);

signal wdata : std_logic_vector (31 downto 0);
signal wdata_mux : std_logic_vector (15 downto 0);

signal rdata : std_logic_vector (15 downto 0);	-- register data read - FDRE
--signal sdto0 : std_logic;

--component opb_xsb300_ak4565
--  Port ( OPB_Clk : in std_logic;
--         OPB_Rst : in std_logic;
--         OPB_ABus : in std_logic_vector (31 downto 0);
--         OPB_BE : in std_logic_vector (3 downto 0);
--         OPB_DBus : in std_logic_vector (31 downto 0);
--         OPB_RNW : in std_logic;
--	 OPB_select : in std_logic;
--	 OPB_seqAddr : in std_logic;
--
--	 UIO_DBus : out std_logic_vector (31 downto 0);
--	 UIO_errAck : out std_logic;
--	 UIO_retry : out std_logic;
--	 UIO_toutSup : out std_logic;
--	 UIO_xferAck : out std_logic;
--
-- 	 Interrupt : out std_logic;

--       AU_CSN_N : out std_logic;
--       AU_BCLK : out std_logic;
--       AU_MCLK : out std_logic;
--       AU_LRCK : out std_logic;
--        AU_SDTI : out std_logic;
--         AU_SDTO0 : in std_logic);
--         end component;


            
component svga
    Port ( clk : in std_logic;
           pix_clk : in std_logic;
           rst : in std_logic;
           video_data : in std_logic_vector(15 downto 0);
           video_addr : out std_logic_vector(19 downto 0);
           video_req : out std_logic;
           vidout_clk : out std_logic;
           vidout_RCR : out std_logic_vector(9 downto 0);
           vidout_GY : out std_logic_vector(9 downto 0);
           vidout_BCB : out std_logic_vector(9 downto 0);
           vidout_BLANK_N : out std_logic;
           vidout_HSYNC_N : out std_logic;
           vidout_VSYNC_N : out std_logic);
end component;

component memoryctrl 
    Port ( rst : in std_logic;
           clk : in std_logic;
           cs : in std_logic;
           select0 : in std_logic;
           rnw : in std_logic;
           vreq : in std_logic;
           onecycle : in std_logic;
           videocycle : out std_logic;
           hihalf : out std_logic;
           pb_wr : out std_logic;
           pb_rd : out std_logic;
           xfer : out std_logic;
           ce0 : out std_logic;
           ce1 : out std_logic;
           rres : out std_logic;
           video_ce : out std_logic);
end component;

component pad_io 
    Port ( clk : in std_logic;
           rst : in std_logic;
           PB_A : out std_logic_vector(19 downto 0);
           PB_UB_N : out std_logic;
           PB_LB_N : out std_logic;
           PB_WE_N : out std_logic;
           PB_OE_N : out std_logic;
           RAM_CE_N : out std_logic;
           PB_D : inout std_logic_vector(15 downto 0);
           pb_addr : in std_logic_vector(19 downto 0);
           pb_ub : in std_logic;
           pb_lb : in std_logic;
           pb_wr : in std_logic;
           pb_rd : in std_logic;
           ram_ce : in std_logic;
           pb_dread : out std_logic_vector(15 downto 0);
           pb_dwrite : in std_logic_vector(15 downto 0));
end component;


begin

--  audio : opb_xsb300_ak4565
--  port map ( OPB_Clk => OPB_Clk,
--         OPB_Rst => OPB_Rst,
--         OPB_ABus => OPB_ABus,
--         OPB_BE => OPB_BE,
--         OPB_DBus => OPB_DBus,
--         OPB_RNW => OPB_RNW,
--	 OPB_select => OPB_select,
--	 OPB_seqAddr => OPB_seqAddr,

--	 UIO_DBus => UIO_DBus,
--	 UIO_errAck => UIO_errAck,
--	 UIO_retry => UIO_retry,
--	 UIO_toutSup => UIO_toutSup,
--	 UIO_xferAck => UIO_xferAck,

	-- Interrupt => Interrupt,

         --AU_CSN_N => AU_CSN_N,
         --AU_BCLK => AU_BLCK,
         --AU_MCLK => AU_MCLK,
         --AU_LRCK => AU_LRCK,
         --AU_SDTI => AU_SDTI
--         AU_SDTO0 => sdto0
--             );



  svga1 : svga
port map (	clk => OPB_Clk,
				pix_clk => pixel_clock,
				rst => OPB_Rst,
				video_addr => video_addr,
				video_data => video_data,
				video_req => video_req,
				VIDOUT_CLK => VIDOUT_CLK,
				VIDOUT_RCR => VIDOUT_RCR, 
				VIDOUT_GY => VIDOUT_GY, 
				VIDOUT_BCB => VIDOUT_BCB,
				VIDOUT_BLANK_N => VIDOUT_BLANK_N,
				VIDOUT_HSYNC_N => VIDOUT_HSYNC_N, 
				VIDOUT_VSYNC_N => VIDOUT_VSYNC_N);

-- the controller state machine
memoryctrl1 : memoryctrl
port map	(	rst => OPB_Rst,
				clk => OPB_Clk,
				cs => cs, 
				select0 => OPB_select, 
				rnw => rnw, 
				vreq => video_req, 
				onecycle => onecycle,
				videocycle => videocycle, 
				hihalf => hihalf, 
				pb_wr => pb_wr, 
				pb_rd => pb_rd, 
				xfer => xfer,
				ce0 => rce0, 
				ce1 => rce1, 
				rres => rreset, 
				video_ce => video_ce);

-- PADS

pad_io1 : pad_io
port map	(	clk => OPB_Clk,
				rst => OPB_Rst,
				PB_A => PB_A,
				PB_UB_N => PB_UB_N,	
				PB_LB_N => PB_LB_N,
				PB_WE_N => PB_WE_N,
				PB_OE_N => PB_OE_N, 
				RAM_CE_N => RAM_CE_N,
				PB_D => PB_D,
				pb_addr => addr_mux,
				pb_rd => pb_rd,
				pb_wr => pb_wr,
				pb_ub => pb_bytesel(1),
				pb_lb => pb_bytesel(0),
				ram_ce => sram_ce,
				pb_dread  => rdata,
				pb_dwrite => wdata_mux);

	sram_ce <= pb_rd or pb_wr;

	amuxsel <= videocycle;

	addr_mux <= video_addr when (amuxsel = '1') else (addr(20 downto 2) & (addr(1) or hihalf));
	--assign addr_mux = amuxsel ? video_addr : {addr[20:2], addr[1] | hihalf};

	onecycle <= (not be(3)) or (not be(2)) or (not be(1)) or (not be(0));

	--assign pb_bytesel = (addr[1] | hihalf ? be[1:0] : be[3:2])& { 2{pb_rd | pb_wr}} | {2 {videocycle}};

	wdata_mux <= wdata(15 downto 0) when ((addr(1) or hihalf) = '1') else wdata(31 downto 16);
	--assign wdata_mux = addr[1] | hihalf ? wdata[15:0] : wdata[31:16];


process(videocycle, be, addr(1), hihalf, pb_rd, pb_wr)
begin

if videocycle='1' then pb_bytesel <= "11";
  elsif pb_rd='1' or pb_wr='1' then
    if addr(1)='1' or hihalf='1' then
      pb_bytesel <= be(1 downto 0);
    else
      pb_bytesel <= be(3 downto 2);
    end if;
  else
    pb_bytesel <= "00";
end if;

  
end process;
      

                      

process (OPB_Clk,	OPB_Rst,	OPB_ABus, OPB_BE, OPB_DBus, 
			OPB_RNW, OPB_select, OPB_seqAddr, pixel_clock)

begin

--	for i in 0 to 1 loop
--		if (addr(1) or hihalf) = '1' then
--			pb_bytesel(i) <= ( be(i) and (pb_rd or pb_wr)) or videocycle; 
--		else 
--			pb_bytesel(i) <= ( be(i + 2) and (pb_rd or pb_wr)) or videocycle;
--		end if;
--	end loop;

	if (OPB_ABus(31 downto 20) = X"008") then
		cs <= OPB_select;
	else
		cs <= '0';
	end if;

-- register rw
	if OPB_Clk'event and OPB_Clk ='1' then
	
		if OPB_Rst = '1' then
			rnw <= '0'; 
		else 
			rnw <= OPB_RNW;
		end if;
	
	end if;

-- register adresses A23 .. A0
	if OPB_Clk'event and OPB_Clk ='1' then
	   for i in 0 to 23 loop
			if OPB_Rst = '1' then
  				addr(i) <= '0';
			else
				addr(i) <= OPB_ABus(i);
			end if;	
   	end loop;
	end if;
			
	-- register BE
	if OPB_Clk'event and OPB_Clk ='1' then
			if OPB_Rst = '1' then
  				be <= "0000";
			else
				be <= OPB_BE;
			end if;
	end if;

-- register data write
	if OPB_Clk'event and OPB_Clk ='1' then
	   for i in 0 to 31 loop
			if OPB_Rst = '1' then
  				wdata(i) <= '0';
			else
				wdata(i) <= OPB_DBus(i);
			end if;		
   	end loop;
	end if;

-- the fun begins
-- ce0/ce1 enables writing MSB (low) / LSB (high) halves

--always @(posedge OPB_Rst or posedge OPB_Clk) begin (synchrounous or asynchronous reset??) 

  		for i in 0 to 15 loop
			if OPB_Rst = '1' then
				UIO_DBus(i) <= '0';
			
			elsif OPB_Clk'event and OPB_Clk ='1' then 
				if rreset = '1' then
					UIO_DBus(i) <= '0';
				elsif (rce1 or rce0) = '1' then
					UIO_DBus(i) <= rdata(i);
				end if;
			end if;
		end loop;

		for i in 16 to 31 loop 
			if OPB_Rst = '1' then
				UIO_DBus(i) <= '0';
			elsif OPB_Clk'event and OPB_Clk ='1' then 

				if rreset = '1' then
					UIO_DBus(i) <= '0';
				elsif rce0 = '1' then
					UIO_DBus(i) <= rdata(i-16);
				end if;
			end if;
		end loop;


	if OPB_Clk'event and OPB_Clk ='1' then
		if	video_ce = '1' then 
			video_data <= rdata;
		end if;
	end if;
-- tie unused to ground
UIO_errAck <= '0';
UIO_retry <= '0';
UIO_toutSup <= '0';

UIO_xferAck <= xfer;

end process;

end Behavioral;
